from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.compute._models.auto_scale import (
    AutoScale,
    AutoScaleParam,
)
from databricks.bundles.compute._models.aws_attributes import (
    AwsAttributes,
    AwsAttributesParam,
)
from databricks.bundles.compute._models.azure_attributes import (
    AzureAttributes,
    AzureAttributesParam,
)
from databricks.bundles.compute._models.cluster_log_conf import (
    ClusterLogConf,
    ClusterLogConfParam,
)
from databricks.bundles.compute._models.data_security_mode import (
    DataSecurityMode,
    DataSecurityModeParam,
)
from databricks.bundles.compute._models.docker_image import (
    DockerImage,
    DockerImageParam,
)
from databricks.bundles.compute._models.gcp_attributes import (
    GcpAttributes,
    GcpAttributesParam,
)
from databricks.bundles.compute._models.init_script_info import (
    InitScriptInfo,
    InitScriptInfoParam,
)
from databricks.bundles.compute._models.runtime_engine import (
    RuntimeEngine,
    RuntimeEngineParam,
)
from databricks.bundles.compute._models.workload_type import (
    WorkloadType,
    WorkloadTypeParam,
)
from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import (
    VariableOrDict,
    VariableOrList,
    VariableOrOptional,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class ClusterSpec:
    """
    Contains a snapshot of the latest user specified settings that were used to create/edit the cluster.
    """

    apply_policy_default_values: VariableOrOptional[bool] = None
    """
    When set to true, fixed and default values from the policy will be used for fields that are omitted. When set to false, only fixed values from the policy will be applied.
    """

    autoscale: VariableOrOptional[AutoScale] = None
    """
    Parameters needed in order to automatically scale clusters up and down based on load.
    Note: autoscaling works best with DB runtime versions 3.0 or later.
    """

    autotermination_minutes: VariableOrOptional[int] = None
    """
    Automatically terminates the cluster after it is inactive for this time in minutes. If not set,
    this cluster will not be automatically terminated. If specified, the threshold must be between
    10 and 10000 minutes.
    Users can also set this value to 0 to explicitly disable automatic termination.
    """

    aws_attributes: VariableOrOptional[AwsAttributes] = None
    """
    Attributes related to clusters running on Amazon Web Services.
    If not specified at cluster creation, a set of default values will be used.
    """

    azure_attributes: VariableOrOptional[AzureAttributes] = None
    """
    Attributes related to clusters running on Microsoft Azure.
    If not specified at cluster creation, a set of default values will be used.
    """

    cluster_log_conf: VariableOrOptional[ClusterLogConf] = None
    """
    The configuration for delivering spark logs to a long-term storage destination.
    Three kinds of destinations (DBFS, S3 and Unity Catalog volumes) are supported. Only one destination can be specified
    for one cluster. If the conf is given, the logs will be delivered to the destination every
    `5 mins`. The destination of driver logs is `$destination/$clusterId/driver`, while
    the destination of executor logs is `$destination/$clusterId/executor`.
    """

    cluster_name: VariableOrOptional[str] = None
    """
    Cluster name requested by the user. This doesn't have to be unique.
    If not specified at creation, the cluster name will be an empty string.
    For job clusters, the cluster name is automatically set based on the job and job run IDs.
    """

    custom_tags: VariableOrDict[str] = field(default_factory=dict)
    """
    Additional tags for cluster resources. Databricks will tag all cluster resources (e.g., AWS
    instances and EBS volumes) with these tags in addition to `default_tags`. Notes:
    
    - Currently, Databricks allows at most 45 custom tags
    
    - Clusters can only reuse cloud resources if the resources' tags are a subset of the cluster tags
    """

    data_security_mode: VariableOrOptional[DataSecurityMode] = None

    docker_image: VariableOrOptional[DockerImage] = None

    driver_instance_pool_id: VariableOrOptional[str] = None
    """
    The optional ID of the instance pool for the driver of the cluster belongs.
    The pool cluster uses the instance pool with id (instance_pool_id) if the driver pool is not
    assigned.
    """

    driver_node_type_id: VariableOrOptional[str] = None
    """
    The node type of the Spark driver.
    Note that this field is optional; if unset, the driver node type will be set as the same value
    as `node_type_id` defined above.
    
    This field, along with node_type_id, should not be set if virtual_cluster_size is set.
    If both driver_node_type_id, node_type_id, and virtual_cluster_size are specified, driver_node_type_id and node_type_id take precedence.
    """

    enable_elastic_disk: VariableOrOptional[bool] = None
    """
    Autoscaling Local Storage: when enabled, this cluster will dynamically acquire additional disk
    space when its Spark workers are running low on disk space. This feature requires specific AWS
    permissions to function correctly - refer to the User Guide for more details.
    """

    enable_local_disk_encryption: VariableOrOptional[bool] = None
    """
    Whether to enable LUKS on cluster VMs' local disks
    """

    gcp_attributes: VariableOrOptional[GcpAttributes] = None
    """
    Attributes related to clusters running on Google Cloud Platform.
    If not specified at cluster creation, a set of default values will be used.
    """

    init_scripts: VariableOrList[InitScriptInfo] = field(default_factory=list)
    """
    The configuration for storing init scripts. Any number of destinations can be specified.
    The scripts are executed sequentially in the order provided.
    If `cluster_log_conf` is specified, init script logs are sent to `<destination>/<cluster-ID>/init_scripts`.
    """

    instance_pool_id: VariableOrOptional[str] = None
    """
    The optional ID of the instance pool to which the cluster belongs.
    """

    is_single_node: VariableOrOptional[bool] = None
    """
    This field can only be used when `kind = CLASSIC_PREVIEW`.
    
    When set to true, Databricks will automatically set single node related `custom_tags`, `spark_conf`, and `num_workers`
    """

    node_type_id: VariableOrOptional[str] = None
    """
    This field encodes, through a single value, the resources available to each of
    the Spark nodes in this cluster. For example, the Spark nodes can be provisioned
    and optimized for memory or compute intensive workloads. A list of available node
    types can be retrieved by using the :method:clusters/listNodeTypes API call.
    """

    num_workers: VariableOrOptional[int] = None
    """
    Number of worker nodes that this cluster should have. A cluster has one Spark Driver
    and `num_workers` Executors for a total of `num_workers` + 1 Spark nodes.
    
    Note: When reading the properties of a cluster, this field reflects the desired number
    of workers rather than the actual current number of workers. For instance, if a cluster
    is resized from 5 to 10 workers, this field will immediately be updated to reflect
    the target size of 10 workers, whereas the workers listed in `spark_info` will gradually
    increase from 5 to 10 as the new nodes are provisioned.
    """

    policy_id: VariableOrOptional[str] = None
    """
    The ID of the cluster policy used to create the cluster if applicable.
    """

    runtime_engine: VariableOrOptional[RuntimeEngine] = None

    single_user_name: VariableOrOptional[str] = None
    """
    Single user name if data_security_mode is `SINGLE_USER`
    """

    spark_conf: VariableOrDict[str] = field(default_factory=dict)
    """
    An object containing a set of optional, user-specified Spark configuration key-value pairs.
    Users can also pass in a string of extra JVM options to the driver and the executors via
    `spark.driver.extraJavaOptions` and `spark.executor.extraJavaOptions` respectively.
    """

    spark_env_vars: VariableOrDict[str] = field(default_factory=dict)
    """
    An object containing a set of optional, user-specified environment variable key-value pairs.
    Please note that key-value pair of the form (X,Y) will be exported as is (i.e.,
    `export X='Y'`) while launching the driver and workers.
    
    In order to specify an additional set of `SPARK_DAEMON_JAVA_OPTS`, we recommend appending
    them to `$SPARK_DAEMON_JAVA_OPTS` as shown in the example below. This ensures that all
    default databricks managed environmental variables are included as well.
    
    Example Spark environment variables:
    `{"SPARK_WORKER_MEMORY": "28000m", "SPARK_LOCAL_DIRS": "/local_disk0"}` or
    `{"SPARK_DAEMON_JAVA_OPTS": "$SPARK_DAEMON_JAVA_OPTS -Dspark.shuffle.service.enabled=true"}`
    """

    spark_version: VariableOrOptional[str] = None
    """
    The Spark version of the cluster, e.g. `3.3.x-scala2.11`.
    A list of available Spark versions can be retrieved by using
    the :method:clusters/sparkVersions API call.
    """

    ssh_public_keys: VariableOrList[str] = field(default_factory=list)
    """
    SSH public key contents that will be added to each Spark node in this cluster. The
    corresponding private keys can be used to login with the user name `ubuntu` on port `2200`.
    Up to 10 keys can be specified.
    """

    use_ml_runtime: VariableOrOptional[bool] = None
    """
    This field can only be used when `kind = CLASSIC_PREVIEW`.
    
    `effective_spark_version` is determined by `spark_version` (DBR release), this field `use_ml_runtime`, and whether `node_type_id` is gpu node or not.
    """

    workload_type: VariableOrOptional[WorkloadType] = None

    @classmethod
    def from_dict(cls, value: "ClusterSpecDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "ClusterSpecDict":
        return _transform_to_json_value(self)  # type:ignore


class ClusterSpecDict(TypedDict, total=False):
    """"""

    apply_policy_default_values: VariableOrOptional[bool]
    """
    When set to true, fixed and default values from the policy will be used for fields that are omitted. When set to false, only fixed values from the policy will be applied.
    """

    autoscale: VariableOrOptional[AutoScaleParam]
    """
    Parameters needed in order to automatically scale clusters up and down based on load.
    Note: autoscaling works best with DB runtime versions 3.0 or later.
    """

    autotermination_minutes: VariableOrOptional[int]
    """
    Automatically terminates the cluster after it is inactive for this time in minutes. If not set,
    this cluster will not be automatically terminated. If specified, the threshold must be between
    10 and 10000 minutes.
    Users can also set this value to 0 to explicitly disable automatic termination.
    """

    aws_attributes: VariableOrOptional[AwsAttributesParam]
    """
    Attributes related to clusters running on Amazon Web Services.
    If not specified at cluster creation, a set of default values will be used.
    """

    azure_attributes: VariableOrOptional[AzureAttributesParam]
    """
    Attributes related to clusters running on Microsoft Azure.
    If not specified at cluster creation, a set of default values will be used.
    """

    cluster_log_conf: VariableOrOptional[ClusterLogConfParam]
    """
    The configuration for delivering spark logs to a long-term storage destination.
    Three kinds of destinations (DBFS, S3 and Unity Catalog volumes) are supported. Only one destination can be specified
    for one cluster. If the conf is given, the logs will be delivered to the destination every
    `5 mins`. The destination of driver logs is `$destination/$clusterId/driver`, while
    the destination of executor logs is `$destination/$clusterId/executor`.
    """

    cluster_name: VariableOrOptional[str]
    """
    Cluster name requested by the user. This doesn't have to be unique.
    If not specified at creation, the cluster name will be an empty string.
    For job clusters, the cluster name is automatically set based on the job and job run IDs.
    """

    custom_tags: VariableOrDict[str]
    """
    Additional tags for cluster resources. Databricks will tag all cluster resources (e.g., AWS
    instances and EBS volumes) with these tags in addition to `default_tags`. Notes:
    
    - Currently, Databricks allows at most 45 custom tags
    
    - Clusters can only reuse cloud resources if the resources' tags are a subset of the cluster tags
    """

    data_security_mode: VariableOrOptional[DataSecurityModeParam]

    docker_image: VariableOrOptional[DockerImageParam]

    driver_instance_pool_id: VariableOrOptional[str]
    """
    The optional ID of the instance pool for the driver of the cluster belongs.
    The pool cluster uses the instance pool with id (instance_pool_id) if the driver pool is not
    assigned.
    """

    driver_node_type_id: VariableOrOptional[str]
    """
    The node type of the Spark driver.
    Note that this field is optional; if unset, the driver node type will be set as the same value
    as `node_type_id` defined above.
    
    This field, along with node_type_id, should not be set if virtual_cluster_size is set.
    If both driver_node_type_id, node_type_id, and virtual_cluster_size are specified, driver_node_type_id and node_type_id take precedence.
    """

    enable_elastic_disk: VariableOrOptional[bool]
    """
    Autoscaling Local Storage: when enabled, this cluster will dynamically acquire additional disk
    space when its Spark workers are running low on disk space. This feature requires specific AWS
    permissions to function correctly - refer to the User Guide for more details.
    """

    enable_local_disk_encryption: VariableOrOptional[bool]
    """
    Whether to enable LUKS on cluster VMs' local disks
    """

    gcp_attributes: VariableOrOptional[GcpAttributesParam]
    """
    Attributes related to clusters running on Google Cloud Platform.
    If not specified at cluster creation, a set of default values will be used.
    """

    init_scripts: VariableOrList[InitScriptInfoParam]
    """
    The configuration for storing init scripts. Any number of destinations can be specified.
    The scripts are executed sequentially in the order provided.
    If `cluster_log_conf` is specified, init script logs are sent to `<destination>/<cluster-ID>/init_scripts`.
    """

    instance_pool_id: VariableOrOptional[str]
    """
    The optional ID of the instance pool to which the cluster belongs.
    """

    is_single_node: VariableOrOptional[bool]
    """
    This field can only be used when `kind = CLASSIC_PREVIEW`.
    
    When set to true, Databricks will automatically set single node related `custom_tags`, `spark_conf`, and `num_workers`
    """

    node_type_id: VariableOrOptional[str]
    """
    This field encodes, through a single value, the resources available to each of
    the Spark nodes in this cluster. For example, the Spark nodes can be provisioned
    and optimized for memory or compute intensive workloads. A list of available node
    types can be retrieved by using the :method:clusters/listNodeTypes API call.
    """

    num_workers: VariableOrOptional[int]
    """
    Number of worker nodes that this cluster should have. A cluster has one Spark Driver
    and `num_workers` Executors for a total of `num_workers` + 1 Spark nodes.
    
    Note: When reading the properties of a cluster, this field reflects the desired number
    of workers rather than the actual current number of workers. For instance, if a cluster
    is resized from 5 to 10 workers, this field will immediately be updated to reflect
    the target size of 10 workers, whereas the workers listed in `spark_info` will gradually
    increase from 5 to 10 as the new nodes are provisioned.
    """

    policy_id: VariableOrOptional[str]
    """
    The ID of the cluster policy used to create the cluster if applicable.
    """

    runtime_engine: VariableOrOptional[RuntimeEngineParam]

    single_user_name: VariableOrOptional[str]
    """
    Single user name if data_security_mode is `SINGLE_USER`
    """

    spark_conf: VariableOrDict[str]
    """
    An object containing a set of optional, user-specified Spark configuration key-value pairs.
    Users can also pass in a string of extra JVM options to the driver and the executors via
    `spark.driver.extraJavaOptions` and `spark.executor.extraJavaOptions` respectively.
    """

    spark_env_vars: VariableOrDict[str]
    """
    An object containing a set of optional, user-specified environment variable key-value pairs.
    Please note that key-value pair of the form (X,Y) will be exported as is (i.e.,
    `export X='Y'`) while launching the driver and workers.
    
    In order to specify an additional set of `SPARK_DAEMON_JAVA_OPTS`, we recommend appending
    them to `$SPARK_DAEMON_JAVA_OPTS` as shown in the example below. This ensures that all
    default databricks managed environmental variables are included as well.
    
    Example Spark environment variables:
    `{"SPARK_WORKER_MEMORY": "28000m", "SPARK_LOCAL_DIRS": "/local_disk0"}` or
    `{"SPARK_DAEMON_JAVA_OPTS": "$SPARK_DAEMON_JAVA_OPTS -Dspark.shuffle.service.enabled=true"}`
    """

    spark_version: VariableOrOptional[str]
    """
    The Spark version of the cluster, e.g. `3.3.x-scala2.11`.
    A list of available Spark versions can be retrieved by using
    the :method:clusters/sparkVersions API call.
    """

    ssh_public_keys: VariableOrList[str]
    """
    SSH public key contents that will be added to each Spark node in this cluster. The
    corresponding private keys can be used to login with the user name `ubuntu` on port `2200`.
    Up to 10 keys can be specified.
    """

    use_ml_runtime: VariableOrOptional[bool]
    """
    This field can only be used when `kind = CLASSIC_PREVIEW`.
    
    `effective_spark_version` is determined by `spark_version` (DBR release), this field `use_ml_runtime`, and whether `node_type_id` is gpu node or not.
    """

    workload_type: VariableOrOptional[WorkloadTypeParam]


ClusterSpecParam = ClusterSpecDict | ClusterSpec



import argparse
import importlib
import inspect
import json
import logging
import os.path
import sys
from copy import deepcopy
from dataclasses import dataclass, field, fields, replace
from typing import (
    Callable,
    Optional,
    TextIO,
)

from databricks.bundles.core import Bundle, Diagnostics, Location, Resources
from databricks.bundles.core._resource_mutator import ResourceMutator
from databricks.bundles.core._resource_type import _ResourceType
from databricks.bundles.core._transform import _transform

__all__ = []


@dataclass
class _Args:
    phase: str
    input: str
    output: str
    diagnostics: str
    locations: Optional[str]  # FIXME should be required
    unknown_args: list[str]


@dataclass
class _Conf:
    resources: list[str] = field(default_factory=list)
    mutators: list[str] = field(default_factory=list)
    venv_path: Optional[str] = None

    @classmethod
    def from_dict(cls, d: dict) -> "_Conf":
        known_keys = [f.name for f in fields(cls)]
        unknown_keys = d.keys() - known_keys

        if unknown_keys:
            logging.warning(f"Unknown configuration keys: {unknown_keys}")

        return _transform(cls, {k: v for k, v in d.items() if k in known_keys})


def _parse_args(args: list[str]) -> _Args:
    parser = argparse.ArgumentParser()
    parser.add_argument("--phase", default=None)
    parser.add_argument("--input", default=None)
    parser.add_argument("--output", default=None)
    parser.add_argument("--diagnostics", default=None)
    parser.add_argument("--locations", default=None)

    parsed, unknown_args = parser.parse_known_args(args)

    if not parsed.phase:
        raise ValueError("Missing required argument --phase")

    if not parsed.input:
        raise ValueError("Missing required argument --input")

    if not parsed.output:
        raise ValueError("Missing required argument --output")

    if not parsed.diagnostics:
        raise ValueError("Missing required argument --diagnostics")

    return _Args(
        phase=parsed.phase,
        input=parsed.input,
        output=parsed.output,
        diagnostics=parsed.diagnostics,
        locations=parsed.locations,
        unknown_args=unknown_args,
    )


def _load_resources_from_input(input: dict) -> tuple[Resources, Diagnostics]:
    resources = Resources()
    diagnostics = Diagnostics()

    input_resources = input.get("resources", {})

    for tpe in _ResourceType.all():
        input_resources_by_tpe = input_resources.get(tpe.plural_name, {})

        for resource_name, resource_dict in input_resources_by_tpe.items():
            try:
                resource = _transform(tpe.resource_type, resource_dict)

                resources.add_resource(resource_name=resource_name, resource=resource)
            except Exception as exc:
                diagnostics = diagnostics.extend(
                    Diagnostics.from_exception(
                        exc=exc,
                        summary=f"Error while loading {tpe.singular_name}",
                        path=("resources", tpe.plural_name, resource_name),
                    )
                )

    return resources, diagnostics


def _apply_mutators(
    bundle: Bundle,
    resources: Resources,
    mutator_functions: list[ResourceMutator],
) -> tuple[Resources, Diagnostics]:
    diagnostics = Diagnostics()

    for tpe in _ResourceType.all():
        resources, diagnostics = diagnostics.extend_tuple(
            _apply_mutators_for_type(bundle, resources, tpe, mutator_functions)
        )

    return resources, diagnostics


def _apply_mutators_for_type(
    bundle: Bundle,
    resources: Resources,
    tpe: _ResourceType,
    mutator_functions: list[ResourceMutator],
) -> tuple[Resources, Diagnostics]:
    resources_dict = getattr(resources, tpe.plural_name)

    for resource_name, resource in resources_dict.items():
        for mutator in mutator_functions:
            if mutator.resource_type != tpe.resource_type:
                continue

            location = Location.from_callable(mutator.function)

            try:
                if _get_num_args(mutator.function) == 1:
                    new_resource = mutator.function(resource)
                else:
                    # defensive copy so that one function doesn't affect another
                    new_resource = mutator.function(deepcopy(bundle), resource)

                # mutating resource in-place works, but we can't tell when it happens,
                # so we only update location if new instance is returned

                if new_resource is not resource:
                    if location:
                        resources.add_location(
                            ("resources", tpe.plural_name, resource_name), location
                        )
                    resources_dict[resource_name] = new_resource
                    resource = new_resource
            except Exception as exc:
                mutator_name = mutator.function.__name__

                return resources, Diagnostics.from_exception(
                    exc=exc,
                    summary=f"Failed to apply '{mutator_name}' mutator",
                    location=location,
                    path=("resources", tpe.plural_name, resource_name),
                )

    return resources, Diagnostics()


def python_mutator(
    args: _Args,
) -> tuple[dict, dict[tuple[str, ...], Location], Diagnostics]:
    input = json.load(open(args.input, encoding="utf-8"))
    experimental = input.get("experimental", {})

    if experimental.get("pydabs", {}) != {}:
        return (
            {},
            {},
            Diagnostics.create_error(
                "'experimental/pydabs' is not supported by 'databricks-bundles', use 'experimental/python' instead",
                detail="",
                location=None,
                path=("experimental", "pydabs"),
            ),
        )

    conf_dict = experimental.get("python", {})
    conf = _transform(_Conf, conf_dict)
    bundle = _parse_bundle_info(input)

    if args.phase == "load_resources":
        resource_functions, diagnostics = _load_functions(conf.resources)
        if diagnostics.has_error():
            return input, {}, diagnostics

        resources, diagnostics = diagnostics.extend_tuple(
            _load_resources(bundle, resource_functions)
        )
        if diagnostics.has_error():
            return input, {}, diagnostics

        new_bundle = _append_resources(input, resources)
        locations = _relativize_locations(resources._locations)

        return new_bundle, locations, diagnostics.extend(resources.diagnostics)
    elif args.phase == "apply_mutators":
        mutator_functions, diagnostics = _load_resource_mutators(conf.mutators)
        if diagnostics.has_error():
            return input, {}, diagnostics

        resources, diagnostics = _load_resources_from_input(input)
        if diagnostics.has_error():
            return input, {}, diagnostics

        resources, diagnostics = diagnostics.extend_tuple(
            _apply_mutators(bundle, resources, mutator_functions)
        )
        if diagnostics.has_error():
            return input, {}, diagnostics

        new_bundle = _append_resources(input, resources)
        locations = _relativize_locations(resources._locations)

        return new_bundle, locations, diagnostics.extend(resources.diagnostics)
    else:
        return input, {}, Diagnostics.create_error(f"Unknown phase: {args.phase}")


def _parse_bundle_info(input: dict) -> Bundle:
    bundle = input.get("bundle", {})
    variables = {k: v.get("value") for k, v in input.get("variables", {}).items()}

    return Bundle(
        target=bundle["target"],
        variables=variables,
    )


def _append_resources(bundle: dict, resources: Resources) -> dict:
    """
    Append resources to input without modifying resources that are already present.
    """

    new_bundle = bundle.copy()

    for tpe in _ResourceType.all():
        resources_dict = getattr(resources, tpe.plural_name)

        if resources_dict:
            new_bundle["resources"] = new_bundle.get("resources", {})
            new_bundle["resources"][tpe.plural_name] = new_bundle["resources"].get(
                tpe.plural_name, {}
            )

            for resource_name, resource in resources_dict.items():
                new_bundle["resources"][tpe.plural_name][resource_name] = (
                    resource.as_dict()
                )

    return new_bundle


def _get_num_args(func: Callable) -> int:
    return len(inspect.signature(func).parameters)


def _load_resources(
    bundle: Bundle,
    functions: list[Callable],
) -> tuple[Resources, Diagnostics]:
    diagnostics = Diagnostics()
    resources = Resources()

    for function in functions:
        try:
            function_resources, diagnostics = diagnostics.extend_tuple(
                _load_resources_from_function(bundle, function)
            )

            resources.add_resources(function_resources)
        except Exception as exc:
            diagnostics = diagnostics.extend(
                Diagnostics.from_exception(
                    exc=exc,
                    summary="Failed to load resources",
                    location=Location.from_callable(function),
                )
            )

    return resources, diagnostics


def _load_functions(names: list[str]) -> tuple[list[Callable], Diagnostics]:
    diagnostics = Diagnostics()
    functions = []

    for name in names:
        try:
            function, diagnostics = diagnostics.extend_tuple(_load_function(name))

            if function:
                functions.append(function)
        except Exception as exc:
            diagnostics = diagnostics.extend(
                Diagnostics.from_exception(
                    exc=exc,
                    summary=f"Failed to load function '{name}'",
                )
            )

    return functions, diagnostics


def _load_resource_mutators(
    names: list[str],
) -> tuple[list[ResourceMutator], Diagnostics]:
    diagnostics = Diagnostics()
    functions = []

    for name in names:
        try:
            function, diagnostics = diagnostics.extend_tuple(
                _load_resource_mutator(name)
            )

            if function:
                functions.append(function)
        except Exception as exc:
            diagnostics = diagnostics.extend(
                Diagnostics.from_exception(
                    exc=exc,
                    summary=f"Failed to load mutator '{name}'",
                )
            )

    return functions, diagnostics


def _load_object(qualified_name: str) -> tuple[Optional[Callable], Diagnostics]:
    diagnostics = Diagnostics()
    [module_name, name] = qualified_name.split(":")

    common_error = qualified_name == "resources:load_resources"

    try:
        module = importlib.import_module(module_name)
    except Exception as exc:
        if isinstance(exc, ModuleNotFoundError) and exc.name == "resources":
            if common_error:
                return None, _explain_common_import_error(exc)

        return None, Diagnostics.from_exception(
            exc=exc,
            summary=f"Failed to import module '{module_name}'",
        )

    try:
        return getattr(module, name), diagnostics
    except AttributeError as exc:
        if common_error:
            return None, _explain_common_import_error(exc)

        return None, Diagnostics.from_exception(
            exc=exc,
            summary=f"Name '{name}' not found in module '{module_name}'",
            location=Location(file=module.__file__) if module.__file__ else None,
        )
    except Exception as exc:
        return None, Diagnostics.from_exception(
            exc=exc,
            summary=f"Failed to load function '{name}' from module '{module_name}'",
        )


def _explain_common_import_error(exc: Exception) -> Diagnostics:
    # a common case when default name of the module is not found
    # we can give a hint to the user how to fix it
    explanation = """Make sure to create a new Python file at resources/__init__.py with contents:

from databricks.bundles.core import load_resources_from_current_package_module


def load_resources():
    return load_resources_from_current_package_module()
"""
    return Diagnostics.from_exception(
        exc=exc,
        summary="Can't find function 'load_resources' in module 'resources'",
        explanation=explanation,
    )


def _load_function(qualified_name: str) -> tuple[Optional[Callable], Diagnostics]:
    [instance, diagnostics] = _load_object(qualified_name)
    [module_name, name] = qualified_name.split(":")

    if diagnostics.has_error():
        return None, diagnostics

    if instance and not isinstance(instance, Callable):
        return None, Diagnostics.create_error(
            f"Function '{name}' in module '{module_name}' is not callable",
        )

    return instance, diagnostics


def _load_resource_mutator(
    qualified_name: str,
) -> tuple[Optional[ResourceMutator], Diagnostics]:
    [instance, diagnostics] = _load_object(qualified_name)
    [module_name, name] = qualified_name.split(":")

    if diagnostics.has_error():
        return None, diagnostics

    if instance and not isinstance(instance, ResourceMutator):
        return None, Diagnostics.create_error(
            f"'{name}' in module '{module_name}' is not instance of ResourceMutator, did you decorate it with @<resource_type>_mutator?",
        )

    return instance, diagnostics


def _load_resources_from_function(
    bundle: Bundle,
    function: Callable,
) -> tuple[Resources, Diagnostics]:
    if _get_num_args(function) == 0:
        resources = function()
    else:
        # defensive copy so that one function doesn't affect another
        resources = function(deepcopy(bundle))

    assert isinstance(resources, Resources)

    return resources, Diagnostics()


def main(argv: list[str]) -> int:
    args = _parse_args(argv[1:])

    if args.unknown_args:
        logging.warning(f"Unknown arguments: {args.unknown_args}")

    logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)
    new_bundle, locations, diagnostics = python_mutator(args)

    with open(args.diagnostics, "w", encoding="utf-8") as f:
        _write_diagnostics(f, diagnostics)

    if locations_path := args.locations:
        with open(locations_path, "w", encoding="utf-8") as f:
            _write_locations(f, locations)

    with open(args.output, "w", encoding="utf-8") as f:
        _write_output(f, new_bundle)

    return 1 if diagnostics.has_error() else 0


def _write_diagnostics(f: TextIO, diagnostics: Diagnostics) -> None:
    for diagnostic in diagnostics.items:
        obj = diagnostic.as_dict()

        if obj.get("path"):
            obj["path"] = ".".join(obj["path"])

        json.dump(obj, f, ensure_ascii=False)
        f.write("\n")


def _write_output(f: TextIO, bundle: dict) -> None:
    json.dump(bundle, f, ensure_ascii=False)


def _relativize_locations(
    locations: dict[tuple[str, ...], Location],
) -> dict[tuple[str, ...], Location]:
    return {
        path: _relativize_location(location) for path, location in locations.items()
    }


def _relativize_location(location: Location) -> Location:
    return replace(location, file=_relativize_path(location.file))


def _relativize_path(path: str) -> str:
    if not os.path.isabs(path):
        return path

    cwd = os.getcwd()
    common = os.path.commonpath([os.getcwd(), path])

    if common == cwd:
        return os.path.relpath(path, cwd)

    return path


def _write_locations(f: TextIO, locations: dict[tuple[str, ...], Location]) -> None:
    for path, location in locations.items():
        obj = {"path": ".".join(path), **location.as_dict()}

        json.dump(obj, f, ensure_ascii=False)
        f.write("\n")


if __name__ == "__main__":
    exit_code = main(sys.argv)

    sys.exit(exit_code)



from enum import Enum
from typing import Literal


class TableSpecificConfigScdType(Enum):
    """
    :meta private: [EXPERIMENTAL]

    The SCD type to use to ingest the table.
    """

    SCD_TYPE_1 = "SCD_TYPE_1"
    SCD_TYPE_2 = "SCD_TYPE_2"


TableSpecificConfigScdTypeParam = (
    Literal["SCD_TYPE_1", "SCD_TYPE_2"] | TableSpecificConfigScdType
)



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.compute._models.adlsgen2_info import (
    Adlsgen2Info,
    Adlsgen2InfoParam,
)
from databricks.bundles.compute._models.gcs_storage_info import (
    GcsStorageInfo,
    GcsStorageInfoParam,
)
from databricks.bundles.compute._models.local_file_info import (
    LocalFileInfo,
    LocalFileInfoParam,
)
from databricks.bundles.compute._models.s3_storage_info import (
    S3StorageInfo,
    S3StorageInfoParam,
)
from databricks.bundles.compute._models.volumes_storage_info import (
    VolumesStorageInfo,
    VolumesStorageInfoParam,
)
from databricks.bundles.compute._models.workspace_storage_info import (
    WorkspaceStorageInfo,
    WorkspaceStorageInfoParam,
)
from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrOptional

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class InitScriptInfo:
    """
    Config for an individual init script
    Next ID: 11
    """

    abfss: VariableOrOptional[Adlsgen2Info] = None
    """
    Contains the Azure Data Lake Storage destination path
    """

    file: VariableOrOptional[LocalFileInfo] = None
    """
    destination needs to be provided, e.g.
    `{ "file": { "destination": "file:/my/local/file.sh" } }`
    """

    gcs: VariableOrOptional[GcsStorageInfo] = None
    """
    destination needs to be provided, e.g.
    `{ "gcs": { "destination": "gs://my-bucket/file.sh" } }`
    """

    s3: VariableOrOptional[S3StorageInfo] = None
    """
    destination and either the region or endpoint need to be provided. e.g.
    `{ \"s3\": { \"destination\": \"s3://cluster_log_bucket/prefix\", \"region\": \"us-west-2\" } }`
    Cluster iam role is used to access s3, please make sure the cluster iam role in
    `instance_profile_arn` has permission to write data to the s3 destination.
    """

    volumes: VariableOrOptional[VolumesStorageInfo] = None
    """
    destination needs to be provided. e.g.
    `{ \"volumes\" : { \"destination\" : \"/Volumes/my-init.sh\" } }`
    """

    workspace: VariableOrOptional[WorkspaceStorageInfo] = None
    """
    destination needs to be provided, e.g.
    `{ "workspace": { "destination": "/cluster-init-scripts/setup-datadog.sh" } }`
    """

    @classmethod
    def from_dict(cls, value: "InitScriptInfoDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "InitScriptInfoDict":
        return _transform_to_json_value(self)  # type:ignore


class InitScriptInfoDict(TypedDict, total=False):
    """"""

    abfss: VariableOrOptional[Adlsgen2InfoParam]
    """
    Contains the Azure Data Lake Storage destination path
    """

    file: VariableOrOptional[LocalFileInfoParam]
    """
    destination needs to be provided, e.g.
    `{ "file": { "destination": "file:/my/local/file.sh" } }`
    """

    gcs: VariableOrOptional[GcsStorageInfoParam]
    """
    destination needs to be provided, e.g.
    `{ "gcs": { "destination": "gs://my-bucket/file.sh" } }`
    """

    s3: VariableOrOptional[S3StorageInfoParam]
    """
    destination and either the region or endpoint need to be provided. e.g.
    `{ \"s3\": { \"destination\": \"s3://cluster_log_bucket/prefix\", \"region\": \"us-west-2\" } }`
    Cluster iam role is used to access s3, please make sure the cluster iam role in
    `instance_profile_arn` has permission to write data to the s3 destination.
    """

    volumes: VariableOrOptional[VolumesStorageInfoParam]
    """
    destination needs to be provided. e.g.
    `{ \"volumes\" : { \"destination\" : \"/Volumes/my-init.sh\" } }`
    """

    workspace: VariableOrOptional[WorkspaceStorageInfoParam]
    """
    destination needs to be provided, e.g.
    `{ "workspace": { "destination": "/cluster-init-scripts/setup-datadog.sh" } }`
    """


InitScriptInfoParam = InitScriptInfoDict | InitScriptInfo



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrList, VariableOrOptional
from databricks.bundles.pipelines._models.ingestion_config import (
    IngestionConfig,
    IngestionConfigParam,
)
from databricks.bundles.pipelines._models.ingestion_source_type import (
    IngestionSourceType,
    IngestionSourceTypeParam,
)
from databricks.bundles.pipelines._models.table_specific_config import (
    TableSpecificConfig,
    TableSpecificConfigParam,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class IngestionPipelineDefinition:
    """"""

    connection_name: VariableOrOptional[str] = None
    """
    Immutable. The Unity Catalog connection that this ingestion pipeline uses to communicate with the source. This is used with connectors for applications like Salesforce, Workday, and so on.
    """

    ingestion_gateway_id: VariableOrOptional[str] = None
    """
    Immutable. Identifier for the gateway that is used by this ingestion pipeline to communicate with the source database. This is used with connectors to databases like SQL Server.
    """

    objects: VariableOrList[IngestionConfig] = field(default_factory=list)
    """
    Required. Settings specifying tables to replicate and the destination for the replicated tables.
    """

    source_type: VariableOrOptional[IngestionSourceType] = None
    """
    The type of the foreign source.
    The source type will be inferred from the source connection or ingestion gateway.
    This field is output only and will be ignored if provided.
    """

    table_configuration: VariableOrOptional[TableSpecificConfig] = None
    """
    Configuration settings to control the ingestion of tables. These settings are applied to all tables in the pipeline.
    """

    @classmethod
    def from_dict(cls, value: "IngestionPipelineDefinitionDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "IngestionPipelineDefinitionDict":
        return _transform_to_json_value(self)  # type:ignore


class IngestionPipelineDefinitionDict(TypedDict, total=False):
    """"""

    connection_name: VariableOrOptional[str]
    """
    Immutable. The Unity Catalog connection that this ingestion pipeline uses to communicate with the source. This is used with connectors for applications like Salesforce, Workday, and so on.
    """

    ingestion_gateway_id: VariableOrOptional[str]
    """
    Immutable. Identifier for the gateway that is used by this ingestion pipeline to communicate with the source database. This is used with connectors to databases like SQL Server.
    """

    objects: VariableOrList[IngestionConfigParam]
    """
    Required. Settings specifying tables to replicate and the destination for the replicated tables.
    """

    source_type: VariableOrOptional[IngestionSourceTypeParam]
    """
    The type of the foreign source.
    The source type will be inferred from the source connection or ingestion gateway.
    This field is output only and will be ignored if provided.
    """

    table_configuration: VariableOrOptional[TableSpecificConfigParam]
    """
    Configuration settings to control the ingestion of tables. These settings are applied to all tables in the pipeline.
    """


IngestionPipelineDefinitionParam = (
    IngestionPipelineDefinitionDict | IngestionPipelineDefinition
)



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.compute._models.maven_library import (
    MavenLibrary,
    MavenLibraryParam,
)
from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrOptional
from databricks.bundles.pipelines._models.file_library import (
    FileLibrary,
    FileLibraryParam,
)
from databricks.bundles.pipelines._models.notebook_library import (
    NotebookLibrary,
    NotebookLibraryParam,
)
from databricks.bundles.pipelines._models.path_pattern import (
    PathPattern,
    PathPatternParam,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class PipelineLibrary:
    """"""

    file: VariableOrOptional[FileLibrary] = None
    """
    The path to a file that defines a pipeline and is stored in the Databricks Repos.
    """

    glob: VariableOrOptional[PathPattern] = None
    """
    :meta private: [EXPERIMENTAL]
    
    The unified field to include source codes.
    Each entry can be a notebook path, a file path, or a folder path that ends `/**`.
    This field cannot be used together with `notebook` or `file`.
    """

    jar: VariableOrOptional[str] = None
    """
    :meta private: [EXPERIMENTAL]
    
    URI of the jar to be installed. Currently only DBFS is supported.
    """

    maven: VariableOrOptional[MavenLibrary] = None
    """
    :meta private: [EXPERIMENTAL]
    
    Specification of a maven library to be installed.
    """

    notebook: VariableOrOptional[NotebookLibrary] = None
    """
    The path to a notebook that defines a pipeline and is stored in the Databricks workspace.
    """

    @classmethod
    def from_dict(cls, value: "PipelineLibraryDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "PipelineLibraryDict":
        return _transform_to_json_value(self)  # type:ignore


class PipelineLibraryDict(TypedDict, total=False):
    """"""

    file: VariableOrOptional[FileLibraryParam]
    """
    The path to a file that defines a pipeline and is stored in the Databricks Repos.
    """

    glob: VariableOrOptional[PathPatternParam]
    """
    :meta private: [EXPERIMENTAL]
    
    The unified field to include source codes.
    Each entry can be a notebook path, a file path, or a folder path that ends `/**`.
    This field cannot be used together with `notebook` or `file`.
    """

    jar: VariableOrOptional[str]
    """
    :meta private: [EXPERIMENTAL]
    
    URI of the jar to be installed. Currently only DBFS is supported.
    """

    maven: VariableOrOptional[MavenLibraryParam]
    """
    :meta private: [EXPERIMENTAL]
    
    Specification of a maven library to be installed.
    """

    notebook: VariableOrOptional[NotebookLibraryParam]
    """
    The path to a notebook that defines a pipeline and is stored in the Databricks workspace.
    """


PipelineLibraryParam = PipelineLibraryDict | PipelineLibrary



from dataclasses import fields, is_dataclass
from enum import Enum
from typing import Any, Union

from databricks.bundles.core._variable import Variable

__all__ = [
    "_transform_to_json_value",
]


def _transform_to_json_value(
    value: Any,
) -> Union[str, bool, int, float, dict[str, Any], list[Any], None]:
    # you should read the returned datatype as recursive JSON

    if isinstance(value, Enum):
        return value.value
    elif isinstance(value, Variable):
        # order matters, because Variable is a dataclass as well
        return value.value
    elif is_dataclass(value):
        return _transform_to_json_object(value)
    elif isinstance(value, list):
        return [_transform_to_json_value(el) for el in value]
    elif isinstance(value, dict):
        return {
            item_key: _transform_to_json_value(item_value)
            for item_key, item_value in value.items()
        }
    elif isinstance(value, Union[str, bool, int, float, None]):
        return value

    raise ValueError(f"Can't serialize '{value}'")


def _transform_to_json_object(value: Any) -> dict[str, Any]:
    # assert is_dataclass(cls) doesn't always work, see:
    # https://github.com/python/cpython/issues/92893

    assert is_dataclass(value), f"Expected dataclass but got {value}"

    out = {}

    for field_name, _ in _stable_fields(value):
        field_value = _transform_to_json_value(getattr(value, field_name))

        # we implement omittempty semantics here skipping None, [], and {}
        if field_value not in [None, [], {}]:
            out[field_name] = field_value

    return out


def _stable_fields(value: Any) -> list[tuple[str, type | str]]:
    assert is_dataclass(value)
    assert not isinstance(value, type)

    field_names = [(field.name, field.type) for field in fields(value)]

    field_names.sort(key=lambda x: x[0])

    return field_names



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOr

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class JobParameterDefinition:
    """"""

    default: VariableOr[str]
    """
    Default value of the parameter.
    """

    name: VariableOr[str]
    """
    The name of the defined parameter. May only contain alphanumeric characters, `_`, `-`, and `.`
    """

    @classmethod
    def from_dict(cls, value: "JobParameterDefinitionDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "JobParameterDefinitionDict":
        return _transform_to_json_value(self)  # type:ignore


class JobParameterDefinitionDict(TypedDict, total=False):
    """"""

    default: VariableOr[str]
    """
    Default value of the parameter.
    """

    name: VariableOr[str]
    """
    The name of the defined parameter. May only contain alphanumeric characters, `_`, `-`, and `.`
    """


JobParameterDefinitionParam = JobParameterDefinitionDict | JobParameterDefinition



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrList

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class TaskEmailNotifications:
    """"""

    on_duration_warning_threshold_exceeded: VariableOrList[str] = field(
        default_factory=list
    )
    """
    A list of email addresses to be notified when the duration of a run exceeds the threshold specified for the `RUN_DURATION_SECONDS` metric in the `health` field. If no rule for the `RUN_DURATION_SECONDS` metric is specified in the `health` field for the job, notifications are not sent.
    """

    on_failure: VariableOrList[str] = field(default_factory=list)
    """
    A list of email addresses to be notified when a run unsuccessfully completes. A run is considered to have completed unsuccessfully if it ends with an `INTERNAL_ERROR` `life_cycle_state` or a `FAILED`, or `TIMED_OUT` result_state. If this is not specified on job creation, reset, or update the list is empty, and notifications are not sent.
    """

    on_start: VariableOrList[str] = field(default_factory=list)
    """
    A list of email addresses to be notified when a run begins. If not specified on job creation, reset, or update, the list is empty, and notifications are not sent.
    """

    on_streaming_backlog_exceeded: VariableOrList[str] = field(default_factory=list)
    """
    A list of email addresses to notify when any streaming backlog thresholds are exceeded for any stream.
    Streaming backlog thresholds can be set in the `health` field using the following metrics: `STREAMING_BACKLOG_BYTES`, `STREAMING_BACKLOG_RECORDS`, `STREAMING_BACKLOG_SECONDS`, or `STREAMING_BACKLOG_FILES`.
    Alerting is based on the 10-minute average of these metrics. If the issue persists, notifications are resent every 30 minutes.
    """

    on_success: VariableOrList[str] = field(default_factory=list)
    """
    A list of email addresses to be notified when a run successfully completes. A run is considered to have completed successfully if it ends with a `TERMINATED` `life_cycle_state` and a `SUCCESS` result_state. If not specified on job creation, reset, or update, the list is empty, and notifications are not sent.
    """

    @classmethod
    def from_dict(cls, value: "TaskEmailNotificationsDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "TaskEmailNotificationsDict":
        return _transform_to_json_value(self)  # type:ignore


class TaskEmailNotificationsDict(TypedDict, total=False):
    """"""

    on_duration_warning_threshold_exceeded: VariableOrList[str]
    """
    A list of email addresses to be notified when the duration of a run exceeds the threshold specified for the `RUN_DURATION_SECONDS` metric in the `health` field. If no rule for the `RUN_DURATION_SECONDS` metric is specified in the `health` field for the job, notifications are not sent.
    """

    on_failure: VariableOrList[str]
    """
    A list of email addresses to be notified when a run unsuccessfully completes. A run is considered to have completed unsuccessfully if it ends with an `INTERNAL_ERROR` `life_cycle_state` or a `FAILED`, or `TIMED_OUT` result_state. If this is not specified on job creation, reset, or update the list is empty, and notifications are not sent.
    """

    on_start: VariableOrList[str]
    """
    A list of email addresses to be notified when a run begins. If not specified on job creation, reset, or update, the list is empty, and notifications are not sent.
    """

    on_streaming_backlog_exceeded: VariableOrList[str]
    """
    A list of email addresses to notify when any streaming backlog thresholds are exceeded for any stream.
    Streaming backlog thresholds can be set in the `health` field using the following metrics: `STREAMING_BACKLOG_BYTES`, `STREAMING_BACKLOG_RECORDS`, `STREAMING_BACKLOG_SECONDS`, or `STREAMING_BACKLOG_FILES`.
    Alerting is based on the 10-minute average of these metrics. If the issue persists, notifications are resent every 30 minutes.
    """

    on_success: VariableOrList[str]
    """
    A list of email addresses to be notified when a run successfully completes. A run is considered to have completed successfully if it ends with a `TERMINATED` `life_cycle_state` and a `SUCCESS` result_state. If not specified on job creation, reset, or update, the list is empty, and notifications are not sent.
    """


TaskEmailNotificationsParam = TaskEmailNotificationsDict | TaskEmailNotifications



from enum import Enum
from typing import Literal


class PipelinePermissionLevel(Enum):
    CAN_MANAGE = "CAN_MANAGE"
    IS_OWNER = "IS_OWNER"
    CAN_RUN = "CAN_RUN"
    CAN_VIEW = "CAN_VIEW"


PipelinePermissionLevelParam = (
    Literal["CAN_MANAGE", "IS_OWNER", "CAN_RUN", "CAN_VIEW"] | PipelinePermissionLevel
)



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOr, VariableOrOptional

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class ComputeConfig:
    """
    :meta private: [EXPERIMENTAL]
    """

    num_gpus: VariableOr[int]
    """
    Number of GPUs.
    """

    gpu_node_pool_id: VariableOrOptional[str] = None
    """
    IDof the GPU pool to use.
    """

    gpu_type: VariableOrOptional[str] = None
    """
    GPU type.
    """

    @classmethod
    def from_dict(cls, value: "ComputeConfigDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "ComputeConfigDict":
        return _transform_to_json_value(self)  # type:ignore


class ComputeConfigDict(TypedDict, total=False):
    """"""

    num_gpus: VariableOr[int]
    """
    Number of GPUs.
    """

    gpu_node_pool_id: VariableOrOptional[str]
    """
    IDof the GPU pool to use.
    """

    gpu_type: VariableOrOptional[str]
    """
    GPU type.
    """


ComputeConfigParam = ComputeConfigDict | ComputeConfig



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOr, VariableOrOptional
from databricks.bundles.jobs._models.compute_config import (
    ComputeConfig,
    ComputeConfigParam,
)
from databricks.bundles.jobs._models.source import Source, SourceParam

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class GenAiComputeTask:
    """
    :meta private: [EXPERIMENTAL]
    """

    dl_runtime_image: VariableOr[str]
    """
    Runtime image
    """

    command: VariableOrOptional[str] = None
    """
    Command launcher to run the actual script, e.g. bash, python etc.
    """

    compute: VariableOrOptional[ComputeConfig] = None

    mlflow_experiment_name: VariableOrOptional[str] = None
    """
    Optional string containing the name of the MLflow experiment to log the run to. If name is not
    found, backend will create the mlflow experiment using the name.
    """

    source: VariableOrOptional[Source] = None
    """
    Optional location type of the training script. When set to `WORKSPACE`, the script will be retrieved from the local Databricks workspace. When set to `GIT`, the script will be retrieved from a Git repository
    defined in `git_source`. If the value is empty, the task will use `GIT` if `git_source` is defined and `WORKSPACE` otherwise.
    * `WORKSPACE`: Script is located in Databricks workspace.
    * `GIT`: Script is located in cloud Git provider.
    """

    training_script_path: VariableOrOptional[str] = None
    """
    The training script file path to be executed. Cloud file URIs (such as dbfs:/, s3:/, adls:/, gcs:/) and workspace paths are supported. For python files stored in the Databricks workspace, the path must be absolute and begin with `/`. For files stored in a remote repository, the path must be relative. This field is required.
    """

    yaml_parameters: VariableOrOptional[str] = None
    """
    Optional string containing model parameters passed to the training script in yaml format.
    If present, then the content in yaml_parameters_file_path will be ignored.
    """

    yaml_parameters_file_path: VariableOrOptional[str] = None
    """
    Optional path to a YAML file containing model parameters passed to the training script.
    """

    @classmethod
    def from_dict(cls, value: "GenAiComputeTaskDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "GenAiComputeTaskDict":
        return _transform_to_json_value(self)  # type:ignore


class GenAiComputeTaskDict(TypedDict, total=False):
    """"""

    dl_runtime_image: VariableOr[str]
    """
    Runtime image
    """

    command: VariableOrOptional[str]
    """
    Command launcher to run the actual script, e.g. bash, python etc.
    """

    compute: VariableOrOptional[ComputeConfigParam]

    mlflow_experiment_name: VariableOrOptional[str]
    """
    Optional string containing the name of the MLflow experiment to log the run to. If name is not
    found, backend will create the mlflow experiment using the name.
    """

    source: VariableOrOptional[SourceParam]
    """
    Optional location type of the training script. When set to `WORKSPACE`, the script will be retrieved from the local Databricks workspace. When set to `GIT`, the script will be retrieved from a Git repository
    defined in `git_source`. If the value is empty, the task will use `GIT` if `git_source` is defined and `WORKSPACE` otherwise.
    * `WORKSPACE`: Script is located in Databricks workspace.
    * `GIT`: Script is located in cloud Git provider.
    """

    training_script_path: VariableOrOptional[str]
    """
    The training script file path to be executed. Cloud file URIs (such as dbfs:/, s3:/, adls:/, gcs:/) and workspace paths are supported. For python files stored in the Databricks workspace, the path must be absolute and begin with `/`. For files stored in a remote repository, the path must be relative. This field is required.
    """

    yaml_parameters: VariableOrOptional[str]
    """
    Optional string containing model parameters passed to the training script in yaml format.
    If present, then the content in yaml_parameters_file_path will be ignored.
    """

    yaml_parameters_file_path: VariableOrOptional[str]
    """
    Optional path to a YAML file containing model parameters passed to the training script.
    """


GenAiComputeTaskParam = GenAiComputeTaskDict | GenAiComputeTask
